@@ -1,3 +1,19 @@
+/*
+ * Copyright 2014 Attila Szegedi, Daniel Dekany, Jonathan Revusky
+ * 
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ * 
+ * http://www.apache.org/licenses/LICENSE-2.0
+ * 
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
+
 package freemarker.ext.beans;
 
 import java.util.AbstractList;
@@ -9,7 +25,6 @@ import freemarker.template.TemplateSequenceModel;
 import freemarker.template.utility.UndeclaredThrowableException;
 
 /**
- * @author Attila Szegedi
  */
 class SequenceAdapter extends AbstractList implements TemplateModelAdapter {
     private final BeansWrapper wrapper;
@@ -41,4 +56,9 @@ class SequenceAdapter extends AbstractList implements TemplateModelAdapter {
             throw new UndeclaredThrowableException(e);
         }
     }
+    
+    public TemplateSequenceModel getTemplateSequenceModel() {
+        return model;
+    }
+    
 }
