@@ -1,53 +1,17 @@
 /*
- * Copyright (c) 2003 The Visigoth Software Society. All rights
- * reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- *
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in
- *    the documentation and/or other materials provided with the
- *    distribution.
- *
- * 3. The end-user documentation included with the redistribution, if
- *    any, must include the following acknowledgement:
- *       "This product includes software developed by the
- *        Visigoth Software Society (http://www.visigoths.org/)."
- *    Alternately, this acknowledgement may appear in the software itself,
- *    if and wherever such third-party acknowledgements normally appear.
- *
- * 4. Neither the name "FreeMarker", "Visigoth", nor any of the names of the
- *    project contributors may be used to endorse or promote products derived
- *    from this software without prior written permission. For written
- *    permission, please contact visigoths@visigoths.org.
- *
- * 5. Products derived from this software may not be called "FreeMarker" or "Visigoth"
- *    nor may "FreeMarker" or "Visigoth" appear in their names
- *    without prior written permission of the Visigoth Software Society.
- *
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED.  IN NO EVENT SHALL THE VISIGOTH SOFTWARE SOCIETY OR
- * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
- * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
- * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
- * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- * ====================================================================
- *
- * This software consists of voluntary contributions made by many
- * individuals on behalf of the Visigoth Software Society. For more
- * information on the Visigoth Software Society, please see
- * http://www.visigoths.org/
+ * Copyright 2014 Attila Szegedi, Daniel Dekany, Jonathan Revusky
+ * 
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ * 
+ * http://www.apache.org/licenses/LICENSE-2.0
+ * 
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
  
 package freemarker.ext.dom;
@@ -80,26 +44,35 @@ import org.xml.sax.ErrorHandler;
 import org.xml.sax.InputSource;
 import org.xml.sax.SAXException;
 
+import freemarker.core._UnexpectedTypeErrorExplainerTemplateModel;
 import freemarker.ext.util.WrapperTemplateModel;
 import freemarker.log.Logger;
 import freemarker.template.AdapterTemplateModel;
+import freemarker.template.DefaultObjectWrapper;
 import freemarker.template.SimpleScalar;
+import freemarker.template.TemplateBooleanModel;
+import freemarker.template.TemplateDateModel;
 import freemarker.template.TemplateHashModel;
 import freemarker.template.TemplateModel;
 import freemarker.template.TemplateModelException;
 import freemarker.template.TemplateNodeModel;
+import freemarker.template.TemplateNumberModel;
 import freemarker.template.TemplateSequenceModel;
 
 /**
  * A base class for wrapping a W3C DOM Node as a FreeMarker template model.
- * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
+ * 
+ * <p>
+ * Note that {@link DefaultObjectWrapper} automatically wraps W3C DOM {@link Node}-s into this, so you may not need to
+ * do that with this class manually. Though, before dropping the {@link Node}-s into the data-model, you may want to
+ * apply {@link NodeModel#simplify(Node)} on them.
  */
 abstract public class NodeModel
 implements TemplateNodeModel, TemplateHashModel, TemplateSequenceModel,
-    AdapterTemplateModel, WrapperTemplateModel
+    AdapterTemplateModel, WrapperTemplateModel, _UnexpectedTypeErrorExplainerTemplateModel
 {
 
-    static final Logger logger = Logger.getLogger("freemarker.dom");
+    static private final Logger LOG = Logger.getLogger("freemarker.dom");
 
     private static final Object STATIC_LOCK = new Object();
     
@@ -119,8 +92,8 @@ implements TemplateNodeModel, TemplateHashModel, TemplateSequenceModel,
         } catch (Exception e) {
             // do nothing
         }
-        if (xpathSupportClass == null && logger.isWarnEnabled()) {
-            logger.warn("No XPath support is available.");
+        if (xpathSupportClass == null && LOG.isWarnEnabled()) {
+            LOG.warn("No XPath support is available.");
         }
     }
     
@@ -583,8 +556,8 @@ implements TemplateNodeModel, TemplateHashModel, TemplateSequenceModel,
         synchronized (STATIC_LOCK) {
             xpathSupportClass = c;
         }
-        if (logger.isDebugEnabled()) {
-            logger.debug("Using Jaxen classes for XPath support");
+        if (LOG.isDebugEnabled()) {
+            LOG.debug("Using Jaxen classes for XPath support");
         }
     }
     
@@ -598,8 +571,8 @@ implements TemplateNodeModel, TemplateHashModel, TemplateSequenceModel,
         synchronized (STATIC_LOCK) {
             xpathSupportClass = c;
         }
-        if (logger.isDebugEnabled()) {
-            logger.debug("Using Xalan classes for XPath support");
+        if (LOG.isDebugEnabled()) {
+            LOG.debug("Using Xalan classes for XPath support");
         }
     }
     
@@ -609,8 +582,8 @@ implements TemplateNodeModel, TemplateHashModel, TemplateSequenceModel,
         synchronized (STATIC_LOCK) {
             xpathSupportClass = c;
         }
-        if (logger.isDebugEnabled()) {
-            logger.debug("Using Sun's internal Xalan classes for XPath support");
+        if (LOG.isDebugEnabled()) {
+            LOG.debug("Using Sun's internal Xalan classes for XPath support");
         }
     }
     
@@ -675,7 +648,7 @@ implements TemplateNodeModel, TemplateHashModel, TemplateSequenceModel,
                     xps = (XPathSupport) xpathSupportClass.newInstance();
                     xpathSupportMap.put(doc, new WeakReference(xps));
                 } catch (Exception e) {
-                    logger.error("Error instantiating xpathSupport class", e);
+                    LOG.error("Error instantiating xpathSupport class", e);
                 }                
             }
         }
@@ -694,4 +667,22 @@ implements TemplateNodeModel, TemplateHashModel, TemplateSequenceModel,
     public Object getWrappedObject() {
         return node;
     }
+    
+    public Object[] explainTypeError(Class[] expectedClasses) {
+        for (int i = 0; i < expectedClasses.length; i++) {
+            Class expectedClass = expectedClasses[i];
+            if (TemplateDateModel.class.isAssignableFrom(expectedClass)
+                    || TemplateNumberModel.class.isAssignableFrom(expectedClass)
+                    || TemplateBooleanModel.class.isAssignableFrom(expectedClass)) {
+                return new Object[] {
+                        "XML node values are always strings (text), that is, they can't be used as number, "
+                        + "date/time/datetime or boolean without explicit conversion (such as "
+                        + "someNode?number, someNode?datetime.xs, someNode?date.xs, someNode?time.xs, "
+                        + "someNode?boolean).",
+                        };
+            }
+        }
+        return null;
+    }
+    
 }
