@@ -1,53 +1,17 @@
 /*
- * Copyright (c) 2003 The Visigoth Software Society. All rights
- * reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- *
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in
- *    the documentation and/or other materials provided with the
- *    distribution.
- *
- * 3. The end-user documentation included with the redistribution, if
- *    any, must include the following acknowledgement:
- *       "This product includes software developed by the
- *        Visigoth Software Society (http://www.visigoths.org/)."
- *    Alternately, this acknowledgement may appear in the software itself,
- *    if and wherever such third-party acknowledgements normally appear.
- *
- * 4. Neither the name "FreeMarker", "Visigoth", nor any of the names of the 
- *    project contributors may be used to endorse or promote products derived
- *    from this software without prior written permission. For written
- *    permission, please contact visigoths@visigoths.org.
- *
- * 5. Products derived from this software may not be called "FreeMarker" or "Visigoth"
- *    nor may "FreeMarker" or "Visigoth" appear in their names
- *    without prior written permission of the Visigoth Software Society.
- *
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED.  IN NO EVENT SHALL THE VISIGOTH SOFTWARE SOCIETY OR
- * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
- * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
- * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
- * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- * ====================================================================
- *
- * This software consists of voluntary contributions made by many
- * individuals on behalf of the Visigoth Software Society. For more
- * information on the Visigoth Software Society, please see
- * http://www.visigoths.org/
+ * Copyright 2014 Attila Szegedi, Daniel Dekany, Jonathan Revusky
+ * 
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ * 
+ * http://www.apache.org/licenses/LICENSE-2.0
+ * 
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
 
 package freemarker.ext.jdom;
@@ -93,7 +57,6 @@ import freemarker.template.TemplateModelException;
 import freemarker.template.TemplateModelIterator;
 import freemarker.template.TemplateScalarModel;
 import freemarker.template.TemplateSequenceModel;
-import freemarker.template.utility.Collections12;
 
 /**
  * Provides a template for wrapping JDOM objects. It is capable of storing not only
@@ -103,7 +66,7 @@ import freemarker.template.utility.Collections12;
  * See individual method documentation for exact details on how the class works. In
  * short:
  * <ul>
- * <li>{@link #getAsString()} will render all contained nodes as XML fragment,</tt>
+ * <li>{@link #getAsString()} will render all contained nodes as XML fragment,
  * <li>{@link #exec(List)} provides full XPath functionality implemented on top of
  * the <a href="http://www.jaxen.org">Jaxen</a> library,</li>
  * <li>{@link #get(String)} provides node traversal, copying and filtering - somewhat
@@ -117,7 +80,6 @@ import freemarker.template.utility.Collections12;
  *   {@link freemarker.ext.xml.NodeListModel freemarker.ext.xml.NodeListModel}
  * 
  * @deprecated Use {@link freemarker.ext.dom.NodeModel} instead.
- * @author Attila Szegedi
  */
 public class NodeListModel
 implements
@@ -157,7 +119,7 @@ implements
      */
     public NodeListModel(Document document)
     {
-        nodes = document == null ? Collections.EMPTY_LIST : Collections12.singletonList(document);
+        nodes = document == null ? Collections.EMPTY_LIST : Collections.singletonList(document);
         namespaces = new HashMap();
     }
 
@@ -166,13 +128,13 @@ implements
      */
     public NodeListModel(Element element)
     {
-        nodes = element == null ? Collections.EMPTY_LIST : Collections12.singletonList(element);
+        nodes = element == null ? Collections.EMPTY_LIST : Collections.singletonList(element);
         namespaces = new HashMap();
     }
 
     private NodeListModel(Object object, Map namespaces)
     {
-        nodes = object == null ? Collections.EMPTY_LIST : Collections12.singletonList(object);
+        nodes = object == null ? Collections.EMPTY_LIST : Collections.singletonList(object);
         this.namespaces = namespaces;
     }
 
@@ -668,7 +630,7 @@ implements
                 return((Element)node).getChildren();
             else if (node instanceof Document) {
                 Element root = ((Document)node).getRootElement();
-                return root == null ? Collections.EMPTY_LIST : Collections12.singletonList(root);
+                return root == null ? Collections.EMPTY_LIST : Collections.singletonList(root);
             } 
  // With 2.1 semantics it  makes more sense to just return a null and let the core 
  // throw an InvalidReferenceException and the template writer can use ?exists etcetera. (JR)
@@ -690,7 +652,7 @@ implements
                 if (root != null &&
                     root.getName().equals(localName) &&
                     root.getNamespaceURI().equals(namespace.getURI())) {
-                    return Collections12.singletonList(root);
+                    return Collections.singletonList(root);
                 } else
                     return Collections.EMPTY_LIST;
             } 
@@ -753,7 +715,7 @@ implements
             else
                 throw new TemplateModelException("_allAttributes can not be applied on " + node.getClass());
 */
-            return attr == null ? Collections.EMPTY_LIST : Collections12.singletonList(attr);
+            return attr == null ? Collections.EMPTY_LIST : Collections.singletonList(attr);
         }
     }
 
@@ -761,15 +723,15 @@ implements
         public List operate(Object node)
         {
             if (node instanceof Element)
-                return Collections12.singletonList(((Element)node).getName());
+                return Collections.singletonList(((Element)node).getName());
             else if (node instanceof Attribute)
-                return Collections12.singletonList(((Attribute)node).getName());
+                return Collections.singletonList(((Attribute)node).getName());
             else if (node instanceof EntityRef)
-                return Collections12.singletonList(((EntityRef)node).getName());
+                return Collections.singletonList(((EntityRef)node).getName());
             else if (node instanceof ProcessingInstruction)
-                return Collections12.singletonList(((ProcessingInstruction)node).getTarget());
+                return Collections.singletonList(((ProcessingInstruction)node).getTarget());
             else if (node instanceof DocType)
-                return Collections12.singletonList(((DocType)node).getPublicID());
+                return Collections.singletonList(((DocType)node).getPublicID());
             else
                 return null;
             // With 2.1 semantics it  makes more sense to just return a null and let the core 
@@ -782,9 +744,9 @@ implements
         public List operate(Object node)
         {
             if (node instanceof Element)
-                return Collections12.singletonList(((Element)node).getQualifiedName());
+                return Collections.singletonList(((Element)node).getQualifiedName());
             else if (node instanceof Attribute)
-                return Collections12.singletonList(((Attribute)node).getQualifiedName());
+                return Collections.singletonList(((Attribute)node).getQualifiedName());
             // With 2.1 semantics it  makes more sense to just return a null and let the core 
             // throw an InvalidReferenceException and the template writer can use ?exists etcetera. (JR)
             return null;
@@ -796,9 +758,9 @@ implements
         public List operate(Object node)
         {
             if (node instanceof Element)
-                return Collections12.singletonList(((Element)node).getNamespace().getURI());
+                return Collections.singletonList(((Element)node).getNamespace().getURI());
             else if (node instanceof Attribute)
-                return Collections12.singletonList(((Attribute)node).getNamespace().getURI());
+                return Collections.singletonList(((Attribute)node).getNamespace().getURI());
             // With 2.1 semantics it  makes more sense to just return a null and let the core 
             // throw an InvalidReferenceException and the template writer can use ?exists etcetera. (JR)
             return null;
@@ -810,9 +772,9 @@ implements
         public List operate(Object node)
         {
             if (node instanceof Element)
-                return Collections12.singletonList(((Element)node).getNamespace().getPrefix());
+                return Collections.singletonList(((Element)node).getNamespace().getPrefix());
             else if (node instanceof Attribute)
-                return Collections12.singletonList(((Attribute)node).getNamespace().getPrefix());
+                return Collections.singletonList(((Attribute)node).getNamespace().getPrefix());
             // With 2.1 semantics it  makes more sense to just return a null and let the core 
             // throw an InvalidReferenceException and the template writer can use ?exists etcetera. (JR)
             return null;
@@ -826,12 +788,12 @@ implements
             if (node instanceof Element)
             {
                 Element element = (Element)node;
-                return Collections12.singletonList(element.getNamespace().getURI() + element.getName());
+                return Collections.singletonList(element.getNamespace().getURI() + element.getName());
             }
             else if (node instanceof Attribute)
             {
                 Attribute attribute = (Attribute)node;
-                return Collections12.singletonList(attribute.getNamespace().getURI() + attribute.getName());
+                return Collections.singletonList(attribute.getNamespace().getURI() + attribute.getName());
             }
             // With 2.1 semantics it  makes more sense to just return a null and let the core 
             // throw an InvalidReferenceException and the template writer can use ?exists etcetera. (JR)
@@ -866,7 +828,7 @@ implements
         public List operate(Object node)
         {
             Element parent = getParent(node);
-            return parent == null ? Collections.EMPTY_LIST : Collections12.singletonList(parent);
+            return parent == null ? Collections.EMPTY_LIST : Collections.singletonList(parent);
         }
     }
 
@@ -889,7 +851,7 @@ implements
         public List operate(Object node)
         {
             Element parent = getParent(node);
-            if (parent == null) return Collections12.singletonList(node);
+            if (parent == null) return Collections.singletonList(node);
             LinkedList list = new LinkedList();
             list.addFirst(node);
             do {
@@ -969,7 +931,7 @@ implements
                 return null;
 //                throw new TemplateModelException("_document can not be applied on " + node.getClass());
 
-            return doc == null ? Collections.EMPTY_LIST : Collections12.singletonList(doc);
+            return doc == null ? Collections.EMPTY_LIST : Collections.singletonList(doc);
         }
     }
 
@@ -978,7 +940,7 @@ implements
         {
             if (node instanceof Document) {
                 DocType doctype = ((Document)node).getDocType();
-                return doctype == null ? Collections.EMPTY_LIST : Collections12.singletonList(doctype);
+                return doctype == null ? Collections.EMPTY_LIST : Collections.singletonList(doctype);
             } else
                 // With 2.1 semantics it  makes more sense to just return a null and let the core 
                 // throw an InvalidReferenceException and the template writer can use ?exists etcetera. (JR)
@@ -1005,15 +967,15 @@ implements
         public List operate(Object node)
         {
             if (node instanceof Element)
-                return Collections12.singletonList(((Element)node).getTextTrim());
+                return Collections.singletonList(((Element)node).getTextTrim());
             if (node instanceof Attribute)
-                return Collections12.singletonList(((Attribute)node).getValue());
+                return Collections.singletonList(((Attribute)node).getValue());
             if (node instanceof CDATA)
-                return Collections12.singletonList(((CDATA)node).getText());
+                return Collections.singletonList(((CDATA)node).getText());
             if (node instanceof Comment)
-                return Collections12.singletonList(((Comment)node).getText());
+                return Collections.singletonList(((Comment)node).getText());
             if (node instanceof ProcessingInstruction)
-                return Collections12.singletonList(((ProcessingInstruction)node).getData());
+                return Collections.singletonList(((ProcessingInstruction)node).getData());
             // With 2.1 semantics it  makes more sense to just return a null and let the core 
             // throw an InvalidReferenceException and the template writer can use ?exists etcetera. (JR)
             return null;
@@ -1240,6 +1202,8 @@ implements
      * document from the standard input, passes it to the template as variable
      * <tt>document</tt> and writes the result of template processing to
      * standard output.
+     * 
+     * @deprecated Will be removed (main method in a library, often classified as CWE-489 "Leftover Debug Code").
      */
     public static void main(String[] args)
     throws
