@@ -1,53 +1,17 @@
 /*
- * Copyright (c) 2003 The Visigoth Software Society. All rights
- * reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- *
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in
- *    the documentation and/or other materials provided with the
- *    distribution.
- *
- * 3. The end-user documentation included with the redistribution, if
- *    any, must include the following acknowledgement:
- *       "This product includes software developed by the
- *        Visigoth Software Society (http://www.visigoths.org/)."
- *    Alternately, this acknowledgement may appear in the software itself,
- *    if and wherever such third-party acknowledgements normally appear.
- *
- * 4. Neither the name "FreeMarker", "Visigoth", nor any of the names of the 
- *    project contributors may be used to endorse or promote products derived
- *    from this software without prior written permission. For written
- *    permission, please contact visigoths@visigoths.org.
- *
- * 5. Products derived from this software may not be called "FreeMarker" or "Visigoth"
- *    nor may "FreeMarker" or "Visigoth" appear in their names
- *    without prior written permission of the Visigoth Software Society.
- *
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED.  IN NO EVENT SHALL THE VISIGOTH SOFTWARE SOCIETY OR
- * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
- * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
- * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
- * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- * ====================================================================
- *
- * This software consists of voluntary contributions made by many
- * individuals on behalf of the Visigoth Software Society. For more
- * information on the Visigoth Software Society, please see
- * http://www.visigoths.org/
+ * Copyright 2014 Attila Szegedi, Daniel Dekany, Jonathan Revusky
+ * 
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ * 
+ * http://www.apache.org/licenses/LICENSE-2.0
+ * 
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
 
 
@@ -85,23 +49,26 @@ import freemarker.ext.servlet.ServletContextHashModel;
 import freemarker.ext.util.WrapperTemplateModel;
 import freemarker.template.AdapterTemplateModel;
 import freemarker.template.ObjectWrapper;
+import freemarker.template.ObjectWrapperAndUnwrapper;
 import freemarker.template.TemplateBooleanModel;
+import freemarker.template.TemplateDateModel;
 import freemarker.template.TemplateHashModelEx;
 import freemarker.template.TemplateModel;
 import freemarker.template.TemplateModelException;
 import freemarker.template.TemplateModelIterator;
 import freemarker.template.TemplateNumberModel;
 import freemarker.template.TemplateScalarModel;
+import freemarker.template._TemplateAPI;
 import freemarker.template.utility.UndeclaredThrowableException;
 
 /**
- * @author Attila Szegedi
  */
 abstract class FreeMarkerPageContext extends PageContext implements TemplateModel
 {
     private static final Class OBJECT_CLASS = Object.class;
         
     private final Environment environment;
+    private final int incompatibleImprovements;
     private List tags = new ArrayList();
     private List outs = new ArrayList();
     private final GenericServlet servlet;
@@ -109,11 +76,13 @@ abstract class FreeMarkerPageContext extends PageContext implements TemplateMode
     private final HttpServletRequest request;
     private final HttpServletResponse response;
     private final ObjectWrapper wrapper;
+    private final ObjectWrapperAndUnwrapper unwrapper;
     private JspWriter jspOut;
     
     protected FreeMarkerPageContext() throws TemplateModelException
     {
         environment = Environment.getCurrentEnvironment();
+        incompatibleImprovements = environment.getConfiguration().getIncompatibleImprovements().intValue();
 
         TemplateModel appModel = environment.getGlobalVariable(
                 FreemarkerServlet.KEY_APPLICATION_PRIVATE);
@@ -140,10 +109,12 @@ abstract class FreeMarkerPageContext extends PageContext implements TemplateMode
         }
         if(requestModel instanceof HttpRequestHashModel) {
             HttpRequestHashModel reqHash = (HttpRequestHashModel)requestModel;
-            this.request = reqHash.getRequest();
-            this.session = request.getSession(false);
-            this.response = reqHash.getResponse();
-            this.wrapper = reqHash.getObjectWrapper();
+            request = reqHash.getRequest();
+            session = request.getSession(false);
+            response = reqHash.getResponse();
+            wrapper = reqHash.getObjectWrapper();
+            unwrapper = this.wrapper instanceof ObjectWrapperAndUnwrapper
+                    ? (ObjectWrapperAndUnwrapper) this.wrapper : null;
         }
         else  {
             throw new  TemplateModelException("Could not find an instance of " + 
@@ -222,26 +193,34 @@ abstract class FreeMarkerPageContext extends PageContext implements TemplateMode
         switch (scope) {
             case PAGE_SCOPE: {
                 try {
-                    TemplateModel m = environment.getGlobalNamespace().get(name);
-                    if (m instanceof AdapterTemplateModel) {
-                        return ((AdapterTemplateModel) m).getAdaptedObject(OBJECT_CLASS);
+                    final TemplateModel tm = environment.getGlobalNamespace().get(name);
+                    if (incompatibleImprovements >= _TemplateAPI.VERSION_INT_2_3_22 && unwrapper != null) {
+                        return unwrapper.unwrap(tm);
+                    } else { // Legacy behavior branch
+                        if (tm instanceof AdapterTemplateModel) {
+                            return ((AdapterTemplateModel) tm).getAdaptedObject(OBJECT_CLASS);
+                        }
+                        if (tm instanceof WrapperTemplateModel) {
+                            return ((WrapperTemplateModel)tm).getWrappedObject();
+                        }
+                        if (tm instanceof TemplateScalarModel) {
+                            return ((TemplateScalarModel) tm).getAsString();
+                        }
+                        if (tm instanceof TemplateNumberModel) {
+                            return ((TemplateNumberModel) tm).getAsNumber();
+                        }
+                        if (tm instanceof TemplateBooleanModel) {
+                            return Boolean.valueOf(((TemplateBooleanModel) tm).getAsBoolean());
+                        }
+                        if (incompatibleImprovements >= _TemplateAPI.VERSION_INT_2_3_22
+                                && tm instanceof TemplateDateModel) {
+                            return ((TemplateDateModel) tm).getAsDate();
+                        }
+                        return tm;
                     }
-                    if (m instanceof WrapperTemplateModel) {
-                        return ((WrapperTemplateModel)m).getWrappedObject();
-                    }
-                    if (m instanceof TemplateScalarModel) {
-                        return ((TemplateScalarModel) m).getAsString();
-                    }
-                    if (m instanceof TemplateNumberModel) {
-                        return ((TemplateNumberModel) m).getAsNumber();
-                    }
-                    if (m instanceof TemplateBooleanModel) {
-                        return ((TemplateBooleanModel) m).getAsBoolean() ? Boolean.TRUE : Boolean.FALSE;
-                    }
-                    return m;
                 }
                 catch (TemplateModelException e) {
-                    throw new UndeclaredThrowableException(e);
+                    throw new UndeclaredThrowableException("Failed to unwrapp FTL global variable", e);
                 }
             }
             case REQUEST_SCOPE: {
