@@ -1,53 +1,17 @@
 /*
- * Copyright (c) 2003 The Visigoth Software Society. All rights
- * reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- *
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in
- *    the documentation and/or other materials provided with the
- *    distribution.
- *
- * 3. The end-user documentation included with the redistribution, if
- *    any, must include the following acknowledgement:
- *       "This product includes software developed by the
- *        Visigoth Software Society (http://www.visigoths.org/)."
- *    Alternately, this acknowledgement may appear in the software itself,
- *    if and wherever such third-party acknowledgements normally appear.
- *
- * 4. Neither the name "FreeMarker", "Visigoth", nor any of the names of the
- *    project contributors may be used to endorse or promote products derived
- *    from this software without prior written permission. For written
- *    permission, please contact visigoths@visigoths.org.
- *
- * 5. Products derived from this software may not be called "FreeMarker" or "Visigoth"
- *    nor may "FreeMarker" or "Visigoth" appear in their names
- *    without prior written permission of the Visigoth Software Society.
- *
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED.  IN NO EVENT SHALL THE VISIGOTH SOFTWARE SOCIETY OR
- * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
- * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
- * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
- * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- * ====================================================================
- *
- * This software consists of voluntary contributions made by many
- * individuals on behalf of the Visigoth Software Society. For more
- * information on the Visigoth Software Society, please see
- * http://www.visigoths.org/
+ * Copyright 2014 Attila Szegedi, Daniel Dekany, Jonathan Revusky
+ * 
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ * 
+ * http://www.apache.org/licenses/LICENSE-2.0
+ * 
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
 
 package freemarker.template;
@@ -70,7 +34,9 @@ import java.util.Vector;
 
 import javax.swing.tree.TreePath;
 
+import freemarker.cache.TemplateCache;
 import freemarker.cache.TemplateLoader;
+import freemarker.cache.TemplateLookupStrategy;
 import freemarker.core.Configurable;
 import freemarker.core.Environment;
 import freemarker.core.FMParser;
@@ -110,88 +76,138 @@ public class Template extends Configurable {
     private List imports = new Vector();
     private TemplateElement rootElement;
     private String encoding, defaultNS;
+    private Object customLookupCondition;
     private int actualTagSyntax;
+    private int actualNamingConvention;
     private final String name;
+    private final String sourceName;
     private final ArrayList lines = new ArrayList();
     private Map prefixToNamespaceURILookup = new HashMap();
     private Map namespaceURIToPrefixLookup = new HashMap();
+    private Version templateLanguageVersion;
 
     /**
      * A prime constructor to which all other constructors should
      * delegate directly or indirectly.
      */
-    private Template(String name, Configuration cfg)
+    private Template(String name, String sourceName, Configuration cfg, boolean overloadSelector)
     {
-        super(cfg != null ? cfg : Configuration.getDefaultConfiguration());
+        super(toNonNull(cfg));
         this.name = name;
+        this.sourceName = sourceName;
+        this.templateLanguageVersion = normalizeTemplateLanguageVersion(toNonNull(cfg).getIncompatibleImprovements());
+    }
+
+    private static Configuration toNonNull(Configuration cfg) {
+        return cfg != null ? cfg : Configuration.getDefaultConfiguration();
     }
-    
 
     /**
-     * Constructs a template from a character stream. Note that this is a relatively expensive operation; where
-     * higher performance matters, you should re-use (cache) {@link Template} instances instead of re-creating them from
-     * the same source again and again.
-     *
-     * @param name the path of the template file relatively to the (virtual) directory that you use to store
-     *        the templates. Shouldn't start with {@code '/'}. Should use {@code '/'}, not {@code '\'}.
-     *        Check {@link #getName()} to see how the name will be used. The name should be independent of the
-     *        actual storage mechanism and physical location as far as possible. Even when the templates are stored
-     *        straightforwardly in real files (they often aren't; see {@link TemplateLoader}), the name shouldn't be an
-     *        absolute file path. Like if the template is stored in {@code "/www/templates/forum/main.ftl"}, and you
-     *        are using {@code "/www/templates/"} as the template root directory via
-     *        {@link Configuration#setDirectoryForTemplateLoading(java.io.File)}, then the template name will be
-     *        {@code "forum/main.ftl"}.
-     * @param reader the character stream to read from. It will always be closed ({@link Reader#close()}).
-     * @param cfg the Configuration object that this Template is associated with.
-     *        If this is {@code null}, the "default" {@link Configuration} object is used,
-     *        which is highly discouraged, because it can easily lead to
-     *        erroneous, unpredictable behavior.
-     *        (See more {@link Configuration#getDefaultConfiguration() here...})
+     * Same as {@link #Template(String, String, Reader, Configuration)} with {@code null} {@code sourceName} parameter.
      */
     public Template(String name, Reader reader, Configuration cfg) throws IOException {
-        this(name, reader, cfg, null);
+        this(name, null, reader, cfg);
     }
-    
+
     /**
      * Convenience constructor for {@link #Template(String, Reader, Configuration)
      * Template(name, new StringReader(reader), cfg)}.
+     * 
+     * @since 2.3.20
      */
     public Template(String name, String sourceCode, Configuration cfg) throws IOException {
-        this(name, new StringReader(sourceCode), cfg, null);
+        this(name, new StringReader(sourceCode), cfg);
     }
+
+    /**
+     * Convenience constructor for {@link #Template(String, String, Reader, Configuration, String) Template(name, null,
+     * reader, cfg, encoding)}.
+     * 
+     * @deprecated In most applications, use {@link #Template(String, Reader, Configuration)} instead, which doesn't
+     *             specify the encoding.
+     */
+    public Template(String name, Reader reader, Configuration cfg, String encoding) throws IOException {
+        this(name, null, reader, cfg, encoding);
+    }
+
+    /**
+     * Constructs a template from a character stream. Note that this is a relatively expensive operation; where higher
+     * performance matters, you should re-use (cache) {@link Template} instances instead of re-creating them from the
+     * same source again and again. ({@link Configuration#getTemplate(String) and its overloads already do such reuse.})
+     * 
+     * @param name
+     *            The path of the template file relatively to the (virtual) directory that you use to store the
+     *            templates (except if {@link #Template(String, String, Reader, Configuration, String) sourceName}
+     *            differs from it). Shouldn't start with {@code '/'}. Should use {@code '/'}, not {@code '\'}. Check
+     *            {@link #getName()} to see how the name will be used. The name should be independent of the actual
+     *            storage mechanism and physical location as far as possible. Even when the templates are stored
+     *            straightforwardly in real files (they often aren't; see {@link TemplateLoader}), the name shouldn't be
+     *            an absolute file path. Like if the template is stored in {@code "/www/templates/forum/main.ftl"}, and
+     *            you are using {@code "/www/templates/"} as the template root directory via
+     *            {@link Configuration#setDirectoryForTemplateLoading(java.io.File)}, then the template name will be
+     *            {@code "forum/main.ftl"}. The name can be {@code null} (should be used for template made on-the-fly
+     *            instead of being loaded from somewhere), in which case relative paths in it will be relative to
+     *            the template root directory (and here again, it's the {@link TemplateLoader} that knows what that
+     *            "physically" means).
+     * @param sourceName
+     *            See {@link #getSourceName()} for the meaning. Can be {@code null}, in which case
+     *            {@link #getSourceName()} will return the same as {@link #getName()}.
+     * @param reader
+     *            The character stream to read from. It will always be closed ({@link Reader#close()}) by this method.
+     * @param cfg
+     *            The Configuration object that this Template is associated with. If this is {@code null}, the "default"
+     *            {@link Configuration} object is used, which is highly discouraged, because it can easily lead to
+     *            erroneous, unpredictable behavior. (See more {@link Configuration#getDefaultConfiguration() here...})
+     * 
+     * @since 2.3.22
+     */
+   public Template(
+           String name, String sourceName, Reader reader, Configuration cfg) throws IOException {
+       this(name, sourceName, reader, cfg, null);
+   }
     
     /**
-     * Same as {@link #Template(String, Reader, Configuration)}, but also specifies the template's encoding.
+     * Same as {@link #Template(String, String, Reader, Configuration)}, but also specifies the template's encoding (not
+     * recommended).
      *
-     * @param encoding This is the encoding that we are supposed to be using. But it's not really necessary because we
-     *        have a {@link Reader} which is already decoded, but it's kept as meta-info. It also has an impact when
-     *        {@code #include}-ing/{@code #import}-ing another template from this template, as its default encoding will
-     *        be this. But this behavior of said directives is considered to be harmful, and will be probably phased
-     *        out. Until that, it's better to leave this on {@code null}, so that the encoding will come from the
-     *        {@link Configuration}. Note that if this is non-{@code null} and there's an {@code #ftl} header with
-     *        encoding, they must match, or else a {@link WrongEncodingException} is thrown. 
-     *        
-     * @deprecated Use {@link #Template(String, Reader, Configuration)} instead.
+     * @param encoding
+     *            This is the encoding that we are supposed to be using. But it's not really necessary because we have a
+     *            {@link Reader} which is already decoded, but it's kept as meta-info. It also has an impact when
+     *            {@code #include}-ing/{@code #import}-ing another template from this template, as its default encoding
+     *            will be this. But this behavior of said directives is considered to be harmful, and will be probably
+     *            phased out. Until that, it's better to leave this on {@code null}, so that the encoding will come from
+     *            the {@link Configuration}. Note that if this is non-{@code null} and there's an {@code #ftl} header
+     *            with encoding, they must match, or else a {@link WrongEncodingException} is thrown.
+     * 
+     * @deprecated In most applications, use {@link #Template(String, String, Reader, Configuration)} instead, which
+     *             doesn't specify the encoding.
+     * 
+     * @since 2.3.22
      */
-    public Template(String name, Reader reader, Configuration cfg, String encoding)
-    throws IOException
-    {
-        this(name, cfg);
+    public Template(
+            String name, String sourceName, Reader reader, Configuration cfg, String encoding) throws IOException {
+        this(name, sourceName, cfg, true);
+        
         this.encoding = encoding;
+        LineTableBuilder ltbReader;
         try {
             if (!(reader instanceof BufferedReader)) {
                 reader = new BufferedReader(reader, 0x1000);
             }
-            reader = new LineTableBuilder(reader);
+            ltbReader = new LineTableBuilder(reader);
+            reader = ltbReader;
             
             try {
+                final Configuration actualCfg = getConfiguration();
                 parser = new FMParser(this, reader,
-                        getConfiguration().getStrictSyntaxMode(),
-                        getConfiguration().getWhitespaceStripping(),
-                        getConfiguration().getTagSyntax(),
-                        getConfiguration().getIncompatibleImprovements().intValue());
+                        actualCfg.getStrictSyntaxMode(),
+                        actualCfg.getWhitespaceStripping(),
+                        actualCfg.getTagSyntax(),
+                        actualCfg.getNamingConvention(),
+                        actualCfg.getIncompatibleImprovements().intValue());
                 this.rootElement = parser.Root();
                 this.actualTagSyntax = parser._getLastTagSyntax();
+                this.actualNamingConvention = parser._getLastNamingConvention();
             }
             catch (TokenMgrError exc) {
                 // TokenMgrError VS ParseException is not an interesting difference for the user, so we just convert it
@@ -203,12 +219,16 @@ public class Template extends Configurable {
             }
         }
         catch(ParseException e) {
-            e.setTemplateName(name);
+            e.setTemplateName(getSourceName());
             throw e;
         }
         finally {
             reader.close();
         }
+        
+        // Throws any exception that JavaCC has silently treated as EOF:
+        ltbReader.throwFailure();
+        
         DebuggerService.registerTemplate(this);
         namespaceURIToPrefixLookup = Collections.unmodifiableMap(namespaceURIToPrefixLookup);
         prefixToNamespaceURILookup = Collections.unmodifiableMap(prefixToNamespaceURILookup);
@@ -223,35 +243,63 @@ public class Template extends Configurable {
      * See more {@link Configuration#getDefaultConfiguration() here...}.
      */
     public Template(String name, Reader reader) throws IOException {
-        this(name, reader, null);
+        this(name, reader, (Configuration) null);
     }
 
     /**
-     * Only used internally.
+     * Only meant to be used internally.
+     * 
+     * @deprecated Has problems setting actualTagSyntax and templateLanguageVersion; will be removed in 2.4.
      */
-    Template(String name, TemplateElement root, Configuration config) {
-        this(name, config);
+    // [2.4] remove this
+    Template(String name, TemplateElement root, Configuration cfg) {
+        this(name, null, cfg, true);
         this.rootElement = root;
         DebuggerService.registerTemplate(this);
     }
-
+    
     /**
-     * Returns a trivial template, one that is just a single block of
-     * plain text, no dynamic content. (Used by the cache module to create
-     * unparsed templates.)
-     * @param name the path of the template file relative to the directory what you use to store
-     *        the templates. See {@link #getName} for more details.
-     * @param content the block of text that this template represents
-     * @param config the configuration to which this template belongs
+     * Same as {@link #getPlainTextTemplate(String, String, String, Configuration)} with {@code null} {@code sourceName}
+     * argument.
      */
     static public Template getPlainTextTemplate(String name, String content, Configuration config) {
-        Template template = new Template(name, config);
-        TextBlock block = new TextBlock(content);
-        template.rootElement = block;
+        return getPlainTextTemplate(name, null, content, config);
+    }
+    
+    /**
+     * Creates a {@link Template} that only contains a single block of static text, no dynamic content.
+     * 
+     * @param name
+     *            See {@link #getName} for more details.
+     * @param sourceName
+     *            See {@link #getSourceName} for more details. If {@code null}, it will be the same as the {@code name}.
+     * @param content
+     *            the block of text that this template represents
+     * @param config
+     *            the configuration to which this template belongs
+     * 
+     * @since 2.3.22
+     */
+    static public Template getPlainTextTemplate(String name, String sourceName, String content, Configuration config) {
+        Template template = new Template(name, sourceName, config, true);
+        template.rootElement = new TextBlock(content);
+        template.actualTagSyntax = config.getTagSyntax();
         DebuggerService.registerTemplate(template);
         return template;
     }
 
+    private static Version normalizeTemplateLanguageVersion(Version incompatibleImprovements) {
+        _TemplateAPI.checkVersionNotNullAndSupported(incompatibleImprovements);
+        int v = incompatibleImprovements.intValue();
+        if (v < _TemplateAPI.VERSION_INT_2_3_19) {
+            return Configuration.VERSION_2_3_0;
+        } else if (v > _TemplateAPI.VERSION_INT_2_3_21) {
+            return Configuration.VERSION_2_3_21;
+        } else { // if 2.3.19 or 2.3.20 or 2.3.21
+            return incompatibleImprovements;
+        }
+    }
+
     /**
      * Executes template, using the data-model provided, writing the generated output
      * to the supplied {@link Writer}.
@@ -415,60 +463,129 @@ public class Template extends Configurable {
     /**
      * The usually path-like (or URL-like) identifier of the template, or possibly {@code null} for non-stored
      * templates. It usually looks like a relative UN*X path; it should use {@code /}, not {@code \}, and shouldn't
-     * start with {@code /} (but there are no hard guarantees). It's not a real path in a file-system, it's just
-     * a name that a {@link TemplateLoader} used to load the backing resource. Or, it can also be a name that was never
-     * used to load the template (directly created with {@link #Template(String, Reader, Configuration)}).
-     * Even if the templates are stored straightforwardly in files, this is relative to the base directory of the
-     * {@link TemplateLoader}. So it really could be anything, except that it has importance in these situations:
+     * start with {@code /} (but there are no hard guarantees). It's not a real path in a file-system, it's just a name
+     * that a {@link TemplateLoader} used to load the backing resource (in simple cases; actually that name is
+     * {@link #getSourceName()}, but see it there). Or, it can also be a name that was never used to load the template
+     * (directly created with {@link #Template(String, Reader, Configuration)}). Even if the templates are stored
+     * straightforwardly in files, this is relative to the base directory of the {@link TemplateLoader}. So it really
+     * could be anything, except that it has importance in these situations:
      * 
-     * <ul>
-     *   <li><p>Relative paths to other templates in this template will be resolved relatively to the directory part of
-     *       this. Like if the template name is {@link "foo/this.ftl"}, then {@code <#include "other.ftl">} gets
-     *       the template with name {@link "foo/other.ftl"}.
-     *   <li><p>It's shown in error messages. So it should be something based on which the user can find the template.
-     *   <li><p>Some tools, like an IDE plugin, uses this to identify (and find) templates.
-     * </ul>
+     * <p>
+     * Relative paths to other templates in this template will be resolved relatively to the directory part of this.
+     * Like if the template name is {@code "foo/this.ftl"}, then {@code <#include "other.ftl">} gets the template with
+     * name {@code "foo/other.ftl"}.
+     * </p>
      * 
-     * <p>Some frameworks use URL-like template names like {@code "someSchema://foo/bar.ftl"}. FreeMarker understands
-     * this notation, so an absolute path like {@code "/baaz.ftl"} in that template will be resolved too
+     * <p>
+     * You should not use this name to indicate error locations, or to find the actual templates in general, because
+     * localized lookup, acquisition and other lookup strategies can transform names before they get to the
+     * {@link TemplateLoader} (the template storage) mechanism. Use {@link #getSourceName()} for these purposes.
+     * </p>
+     * 
+     * <p>
+     * Some frameworks use URL-like template names like {@code "someSchema://foo/bar.ftl"}. FreeMarker understands this
+     * notation, so an absolute path like {@code "/baaz.ftl"} in that template will be resolved too
      * {@code "someSchema://baaz.ftl"}.
      */
     public String getName() {
         return name;
     }
 
+    /**
+     * The name that was actually used to load this template from the {@link TemplateLoader} (or from other custom
+     * storage mechanism). This is what should be shown in error messages as the error location. This is usually the
+     * same as {@link #getName()}, except when localized lookup, template acquisition ({@code *} step in the name), or
+     * other {@link TemplateLookupStrategy} transforms the requested name ({@link #getName()}) to a different final
+     * {@link TemplateLoader}-level name. For example, when you get a template with name {@code "foo.ftl"} then because
+     * of localized lookup, it's possible that something like {@code "foo_en.ftl"} will be loaded behind the scenes.
+     * While the template name will be still the same as the requested template name ({@code "foo.ftl"}), errors should
+     * point to {@code "foo_de.ftl"}. Note that relative paths are always resolved relatively to the {@code name}, not
+     * to the {@code sourceName}.
+     * 
+     * @since 2.3.22
+     */
+    public String getSourceName() {
+        return sourceName != null ? sourceName : getName();
+    }
+
     /**
      * Returns the Configuration object associated with this template.
      */
     public Configuration getConfiguration() {
         return (Configuration) getParent();
     }
+    
+    /**
+     * Return the template language (FTL) version used by this template.
+     * For now (2.3.21) this is the same as {@link Configuration#getIncompatibleImprovements()}, except
+     * that it's normalized to the lowest version where the template language was changed.
+     */
+    Version getTemplateLanguageVersion() {
+        return templateLanguageVersion;
+    }
 
     /**
-     * Sets the character encoding to use for
-     * included files. Usually you don't set this value manually,
-     * instead it's assigned to the template upon loading.
+     * @deprecated Should only be used internally, and might will be removed later.
      */
-
     public void setEncoding(String encoding) {
         this.encoding = encoding;
     }
 
     /**
-     * Returns the character encoding used for reading included files.
+     * Returns the default character encoding used for reading included files.
      */
     public String getEncoding() {
         return this.encoding;
     }
+    
+    /**
+     * Gets the custom lookup condition with which this template was found. See the {@code customLookupCondition}
+     * parameter of {@link Configuration#getTemplate(String, java.util.Locale, Object, String, boolean, boolean)} for
+     * more explanation.
+     * 
+     * @since 2.3.22
+     */
+    public Object getCustomLookupCondition() {
+        return customLookupCondition;
+    }
 
     /**
-     * Returns the tag syntax the parser has chosen for this template: {@link Configuration#SQUARE_BRACKET_TAG_SYNTAX}
-     * or {@link Configuration#ANGLE_BRACKET_TAG_SYNTAX}. If the syntax couldn't be determined (like because there was
-     * no tags in the template), this returns whatever the default is in the current configuration.
+     * Mostly only used internally; setter pair of {@link #getCustomLookupCondition()}. This meant to be called directly
+     * after instantiating the template with its constructor, after a successfull lookup that used this condition. So
+     * this should only be called from code that deals with creating new {@code Template} objects, like from
+     * {@link TemplateCache}.
+     * 
+     * @since 2.3.22
+     */
+    public void setCustomLookupCondition(Object customLookupCondition) {
+        this.customLookupCondition = customLookupCondition;
+    }
+
+    /**
+     * Returns the tag syntax the parser has chosen for this template. If the syntax could be determined, it's
+     * {@link Configuration#SQUARE_BRACKET_TAG_SYNTAX} or {@link Configuration#ANGLE_BRACKET_TAG_SYNTAX}. If the syntax
+     * couldn't be determined (like because there was no tags in the template, or it was a plain text template), this
+     * returns whatever the default is in the current configuration, so it's maybe
+     * {@link Configuration#AUTO_DETECT_TAG_SYNTAX}.
+     * 
+     * @since 2.3.20
      */
     public int getActualTagSyntax() {
         return actualTagSyntax;
     }
+    
+    /**
+     * Returns the naming convention the parser has chosen for this template. If it could be determined, it's
+     * {@link Configuration#LEGACY_NAMING_CONVENTION} or {@link Configuration#CAMEL_CASE_NAMING_CONVENTION}. If it
+     * couldn't be determined (like because there no identifier that's part of the template language was used where
+     * the naming convention matters), this returns whatever the default is in the current configuration, so it's maybe
+     * {@link Configuration#AUTO_DETECT_TAG_SYNTAX}.
+     * 
+     * @since 2.3.23
+     */
+    public int getActualNamingConvention() {
+        return actualNamingConvention;
+    }
 
     /**
      * Dump the raw template in canonical form.
@@ -485,16 +602,18 @@ public class Template extends Configurable {
     }
 
     /**
-     * Called by code internally to maintain
-     * a table of macros
+     * Called by code internally to maintain a table of macros
+     * 
+     * @deprecated Should only be used internally, and might will be removed later.
      */
     public void addMacro(Macro macro) {
         macros.put(macro.getName(), macro);
     }
 
     /**
-     * Called by code internally to maintain
-     * a list of imports
+     * Called by code internally to maintain a list of imports
+     * 
+     * @deprecated Should only be used internally, and might will be removed later.
      */
     public void addImport(LibraryLoad ll) {
         imports.add(ll);
@@ -534,13 +653,17 @@ public class Template extends Configurable {
     }
 
     /**
-     * This is a helper class that builds up the line table
-     * info for us.
+     * Reader that builds up the line table info for us, and also helps in working around JavaCC's exception
+     * suppression.
      */
     private class LineTableBuilder extends FilterReader {
-
-        StringBuffer lineBuf = new StringBuffer();
+        
+        private final StringBuffer lineBuf = new StringBuffer();
         int lastChar;
+        boolean closed;
+        
+        /** Needed to work around JavaCC behavior where it silently treats any errors as EOF. */ 
+        private IOException failure; 
 
         /**
          * @param r the character stream to wrap
@@ -549,19 +672,41 @@ public class Template extends Configurable {
             super(r);
         }
 
+        public void throwFailure() throws IOException {
+            if (failure != null) {
+                throw failure;
+            }
+        }
+
         public int read() throws IOException {
-            int c = in.read();
-            handleChar(c);
-            return c;
+            try {
+                int c = in.read();
+                handleChar(c);
+                return c;
+            } catch (IOException e) {
+                throw rememberException(e);
+            }
+        }
+
+        private IOException rememberException(IOException e) throws IOException {
+            // JavaCC used to read from the Reader after it was closed. So we must not treat that as a failure. 
+            if (!closed) {
+                failure = e;
+            }
+            return e;
         }
 
         public int read(char cbuf[], int off, int len) throws IOException {
-            int numchars = in.read(cbuf, off, len);
-            for (int i=off; i < off+numchars; i++) {
-                char c = cbuf[i];
-                handleChar(c);
+            try {
+                int numchars = in.read(cbuf, off, len);
+                for (int i=off; i < off+numchars; i++) {
+                    char c = cbuf[i];
+                    handleChar(c);
+                }
+                return numchars;
+            } catch (IOException e) {
+                throw rememberException(e);
             }
-            return numchars;
         }
 
         public void close() throws IOException {
@@ -570,6 +715,7 @@ public class Template extends Configurable {
                 lineBuf.setLength(0);
             }
             super.close();
+            closed = true;
         }
 
         private void handleChar(int c) {
@@ -598,22 +744,30 @@ public class Template extends Configurable {
     }
 
     /**
-     *  @return the root TemplateElement object.
+     * @deprecated Should only be used internally, and might will be removed later.
      */
     public TemplateElement getRootTreeNode() {
         return rootElement;
     }
     
+    /**
+     * @deprecated Should only be used internally, and might will be removed later.
+     */
     public Map getMacros() {
         return macros;
     }
 
+    /**
+     * @deprecated Should only be used internally, and might will be removed later.
+     */
     public List getImports() {
         return imports;
     }
 
     /**
      * This is used internally.
+     * 
+     * @deprecated Should only be used internally, and might will be removed later.
      */
     public void addPrefixNSMapping(String prefix, String nsURI) {
         if (nsURI.length() == 0) {
@@ -693,15 +847,13 @@ public class Template extends Configurable {
     }
     
     /**
-     * @return an array of the elements containing the given column and line numbers.
-     * @param column the column
-     * @param line the line
+     * @return an array of the {@link TemplateElement}s containing the given column and line numbers.
+     * @deprecated Should only be used internally, and might will be removed later.
      */
     public TreePath containingElements(int column, int line) {
-        ArrayList elements = new ArrayList();
+        final ArrayList elements = new ArrayList();
         TemplateElement element = rootElement;
-mainloop:
-        while (element.contains(column, line)) {
+        mainloop: while (element.contains(column, line)) {
             elements.add(element);
             for (Enumeration enumeration = element.children(); enumeration.hasMoreElements();) {
                 TemplateElement elem = (TemplateElement) enumeration.nextElement();
@@ -712,18 +864,57 @@ mainloop:
             }
             break;
         }
-        if (elements == null || elements.isEmpty()) {
+        if (elements.isEmpty()) {
             return null;
         }
         return new TreePath(elements.toArray());
     }
 
+    /**
+     * Thrown by the {@link Template} constructors that specify a non-{@code null} encoding whoch doesn't match the
+     * encoding specified in the {@code #ftl} header of the template.
+     */
     static public class WrongEncodingException extends ParseException {
+        private static final long serialVersionUID = 1L;
 
+        /** @deprecated Use {@link #getTemplateSpecifiedEncoding()} instead. */
         public String specifiedEncoding;
+        
+        private final String constructorSpecifiedEncoding;
 
-        public WrongEncodingException(String specifiedEncoding) {
-            this.specifiedEncoding = specifiedEncoding;
+        /**
+         * @deprecated Use {@link #WrongEncodingException(String, String)}.
+         */
+        public WrongEncodingException(String templateSpecifiedEncoding) {
+            this(templateSpecifiedEncoding, null);
+        }
+
+        /**
+         * @since 2.3.22
+         */
+        public WrongEncodingException(String templateSpecifiedEncoding, String constructorSpecifiedEncoding) {
+            this.specifiedEncoding = templateSpecifiedEncoding;
+            this.constructorSpecifiedEncoding = constructorSpecifiedEncoding;
+        }
+        
+        public String getMessage() {
+            return "Encoding specified inside the template (" + specifiedEncoding
+                    + ") doesn't match the encoding specified for the Template constructor"
+                    + (constructorSpecifiedEncoding != null ? " (" + constructorSpecifiedEncoding + ")." : ".");
+        }
+
+        /**
+         * @since 2.3.22
+         */
+        public String getTemplateSpecifiedEncoding() {
+            return specifiedEncoding;
+        }
+
+        /**
+         * @since 2.3.22
+         */
+        public String getConstructorSpecifiedEncoding() {
+            return constructorSpecifiedEncoding;
         }
 
     }
