@@ -1,53 +1,17 @@
 /*
- * Copyright (c) 2005 The Visigoth Software Society. All rights
- * reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- *
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in
- *    the documentation and/or other materials provided with the
- *    distribution.
- *
- * 3. The end-user documentation included with the redistribution, if
- *    any, must include the following acknowledgement:
- *       "This product includes software developed by the
- *        Visigoth Software Society (http://www.visigoths.org/)."
- *    Alternately, this acknowledgement may appear in the software itself,
- *    if and wherever such third-party acknowledgements normally appear.
- *
- * 4. Neither the name "FreeMarker", "Visigoth", nor any of the names of the 
- *    project contributors may be used to endorse or promote products derived
- *    from this software without prior written permission. For written
- *    permission, please contact visigoths@visigoths.org.
- *
- * 5. Products derived from this software may not be called "FreeMarker" or "Visigoth"
- *    nor may "FreeMarker" or "Visigoth" appear in their names
- *    without prior written permission of the Visigoth Software Society.
- *
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
- * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED.  IN NO EVENT SHALL THE VISIGOTH SOFTWARE SOCIETY OR
- * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
- * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
- * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
- * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
- * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- * ====================================================================
- *
- * This software consists of voluntary contributions made by many
- * individuals on behalf of the Visigoth Software Society. For more
- * information on the Visigoth Software Society, please see
- * http://www.visigoths.org/
+ * Copyright 2014 Attila Szegedi, Daniel Dekany, Jonathan Revusky
+ * 
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ * 
+ * http://www.apache.org/licenses/LICENSE-2.0
+ * 
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
 
 package freemarker.test.templatesuite;
@@ -60,8 +24,10 @@ import java.math.BigDecimal;
 import java.math.BigInteger;
 import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.Date;
 import java.util.GregorianCalendar;
 import java.util.HashMap;
+import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
 import java.util.ResourceBundle;
@@ -73,13 +39,24 @@ import java.util.TreeSet;
 import javax.xml.parsers.DocumentBuilder;
 import javax.xml.parsers.DocumentBuilderFactory;
 
+import junit.framework.AssertionFailedError;
+
+import org.junit.Ignore;
 import org.xml.sax.InputSource;
 
+import com.google.common.collect.ImmutableList;
+import com.google.common.collect.ImmutableMap;
+import com.google.common.collect.ImmutableSet;
+
+import freemarker.core.ASTPrinter;
 import freemarker.ext.beans.BeansWrapper;
 import freemarker.ext.beans.BooleanModel;
+import freemarker.ext.beans.Java7MembersOnlyBeansWrapper;
 import freemarker.ext.beans.ResourceBundleModel;
 import freemarker.ext.dom.NodeModel;
 import freemarker.template.Configuration;
+import freemarker.template.DefaultNonListCollectionAdapter;
+import freemarker.template.DefaultObjectWrapper;
 import freemarker.template.SimpleCollection;
 import freemarker.template.SimpleDate;
 import freemarker.template.SimpleNumber;
@@ -91,6 +68,9 @@ import freemarker.template.TemplateMethodModel;
 import freemarker.template.TemplateNodeModel;
 import freemarker.template.TemplateScalarModel;
 import freemarker.template.TemplateSequenceModel;
+import freemarker.template.Version;
+import freemarker.template._TemplateAPI;
+import freemarker.template.utility.NullArgumentException;
 import freemarker.template.utility.NullWriter;
 import freemarker.template.utility.StringUtil;
 import freemarker.test.templatesuite.models.BooleanAndStringTemplateModel;
@@ -99,34 +79,62 @@ import freemarker.test.templatesuite.models.BooleanHash2;
 import freemarker.test.templatesuite.models.BooleanList1;
 import freemarker.test.templatesuite.models.BooleanList2;
 import freemarker.test.templatesuite.models.BooleanVsStringMethods;
+import freemarker.test.templatesuite.models.JavaObjectInfo;
+import freemarker.test.templatesuite.models.Listables;
 import freemarker.test.templatesuite.models.MultiModel1;
 import freemarker.test.templatesuite.models.OverloadedMethods;
+import freemarker.test.templatesuite.models.OverloadedMethods2;
 import freemarker.test.templatesuite.models.VarArgTestModel;
 import freemarker.test.utility.AssertDirective;
 import freemarker.test.utility.AssertEqualsDirective;
 import freemarker.test.utility.AssertFailsDirective;
 import freemarker.test.utility.FileTestCase;
+import freemarker.test.utility.NoOutputDirective;
 
-
+/**
+ * Instances of this are created and called by {@link TemplateTestSuite}. (It's on "Ignore" so that Eclipse doesn't try
+ * to run this alone.) 
+ */
+@Ignore
 public class TemplateTestCase extends FileTestCase {
     
-    private Template template;
-    private HashMap dataModel = new HashMap();
+    // Name of variables exposed to all test FTL-s:
+    private static final String ICI_INT_VALUE_VAR_NAME = "iciIntValue";
+    private static final String TEST_NAME_VAR_NAME = "testName";
+    private static final String JAVA_OBJECT_INFO_VAR_NAME = "javaObjectInfo";
+    private static final String NO_OUTPUT_VAR_NAME = "noOutput";
+    private static final String ASSERT_FAILS_VAR_NAME = "assertFails";
+    private static final String ASSERT_EQUALS_VAR_NAME = "assertEquals";
+    private static final String ASSERT_VAR_NAME = "assert";
     
+    private final String simpleTestName;
     private final String templateName;
     private final String expectedFileName;
     private final boolean noOutput;
     
-    private Configuration conf = new Configuration();
+    private final Configuration conf;
+    private final HashMap dataModel = new HashMap();
     
-    public TemplateTestCase(String name, String templateName, String expectedFileName, boolean noOutput) {
-        super(name);
+    public TemplateTestCase(String testName, String simpleTestName, String templateName, String expectedFileName, boolean noOutput,
+            Version incompatibleImprovements) {
+        super(testName);
+        NullArgumentException.check("testName", testName);
+        
+        NullArgumentException.check("simpleTestName", simpleTestName);
+        this.simpleTestName = simpleTestName;
+        
+        NullArgumentException.check("templateName", templateName);
         this.templateName = templateName;
+        
+        NullArgumentException.check("expectedFileName", expectedFileName);
         this.expectedFileName = expectedFileName;
+        
         this.noOutput = noOutput;
+        
+        conf = new Configuration(incompatibleImprovements);
     }
     
-    public void setConfigParam(String param, String value) throws IOException {
+    public void setSetting(String param, String value) throws IOException {
         if ("auto_import".equals(param)) {
             StringTokenizer st = new StringTokenizer(value);
             if (!st.hasMoreTokens()) fail("Expecting libname");
@@ -137,16 +145,14 @@ public class TemplateTestCase extends FileTestCase {
             if (!st.hasMoreTokens()) fail("Expecting alias after 'as' in autoimport");
             String alias = st.nextToken();
             conf.addAutoImport(alias, libname);
-        }
-        else if ("clear_encoding_map".equals(param)) {
+        } else if ("clear_encoding_map".equals(param)) {
             if (StringUtil.getYesNo(value)) {
                 conf.clearEncodingMap();
             }
-        }
-        else if ("input_encoding".equals(param)) {
+        } else if ("input_encoding".equals(param)) {
             conf.setDefaultEncoding(value);
-        }
-        else {
+        // INCOMPATIBLE_IMPROVEMENTS is a list here, and was already set in the constructor.
+        } else if (!Configuration.INCOMPATIBLE_IMPROVEMENTS.equals(param)) {
             try {
                 conf.setSetting(param, value);
             } catch (TemplateException e) {
@@ -165,24 +171,34 @@ public class TemplateTestCase extends FileTestCase {
      * We really might as well just expose pretty much 
      * the same tree to all our tests. (JR)
      */
+    @SuppressWarnings("boxing")
     public void setUp() throws Exception {
         conf.setDirectoryForTemplateLoading(new File(getTestClassDirectory(), "templates"));
         
-        dataModel.put("assert", AssertDirective.INSTANCE);
-        dataModel.put("assertEquals", AssertEqualsDirective.INSTANCE);
-        dataModel.put("assertFails", AssertFailsDirective.INSTANCE);
+        dataModel.put(ASSERT_VAR_NAME, AssertDirective.INSTANCE);
+        dataModel.put(ASSERT_EQUALS_VAR_NAME, AssertEqualsDirective.INSTANCE);
+        dataModel.put(ASSERT_FAILS_VAR_NAME, AssertFailsDirective.INSTANCE);
+        dataModel.put(NO_OUTPUT_VAR_NAME, NoOutputDirective.INSTANCE);
+
+        dataModel.put(JAVA_OBJECT_INFO_VAR_NAME, JavaObjectInfo.INSTANCE);
+        dataModel.put(TEST_NAME_VAR_NAME, simpleTestName);
+        dataModel.put(ICI_INT_VALUE_VAR_NAME, conf.getIncompatibleImprovements().intValue());
         
         dataModel.put("message", "Hello, world!");
-        
-        final String testName = getName();
-        if (testName.equals("bean-maps")) {
-            BeansWrapper w1 = new BeansWrapper();
-            BeansWrapper w2 = new BeansWrapper();
-            BeansWrapper w3 = new BeansWrapper();
-            BeansWrapper w4 = new BeansWrapper();
-            BeansWrapper w5 = new BeansWrapper();
-            BeansWrapper w6 = new BeansWrapper();
-            BeansWrapper w7 = new BeansWrapper();
+
+        if (simpleTestName.startsWith("api-builtin")) {
+            dataModel.put("map", ImmutableMap.of(1, "a", 2, "b", 3, "c"));
+            dataModel.put("list", ImmutableList.of(1, 2, 3));
+            dataModel.put("set", ImmutableSet.of("a", "b", "c"));
+            dataModel.put("s", "test");
+        } else if (simpleTestName.equals("bean-maps")) {
+            BeansWrapper w1 = new Java7MembersOnlyBeansWrapper();
+            BeansWrapper w2 = new Java7MembersOnlyBeansWrapper();
+            BeansWrapper w3 = new Java7MembersOnlyBeansWrapper();
+            BeansWrapper w4 = new Java7MembersOnlyBeansWrapper();
+            BeansWrapper w5 = new Java7MembersOnlyBeansWrapper();
+            BeansWrapper w6 = new Java7MembersOnlyBeansWrapper();
+            BeansWrapper w7 = new Java7MembersOnlyBeansWrapper();
     
             w1.setExposureLevel(BeansWrapper.EXPOSE_PROPERTIES_ONLY);
             w2.setExposureLevel(BeansWrapper.EXPOSE_PROPERTIES_ONLY);
@@ -200,7 +216,7 @@ public class TemplateTestCase extends FileTestCase {
     
             w7.setSimpleMapWrapper(true);
     
-            Object test = getTestBean();
+            Object test = getTestMapBean();
     
             dataModel.put("m1", w1.wrap(test));
             dataModel.put("m2", w2.wrap(test));
@@ -216,7 +232,7 @@ public class TemplateTestCase extends FileTestCase {
             dataModel.put("s4", w5.wrap("world"));
         }
         
-        else if (testName.equals("beans")) {
+        else if (simpleTestName.equals("beans")) {
             dataModel.put("array", new String[] { "array-0", "array-1"});
             dataModel.put("list", Arrays.asList(new String[] { "list-0", "list-1", "list-2"}));
             Map tmap = new HashMap();
@@ -232,7 +248,7 @@ public class TemplateTestCase extends FileTestCase {
             dataModel.put("enums", BeansWrapper.getDefaultInstance().getEnumModels());
         }
         
-        else if (testName.equals("boolean")) {
+        else if (simpleTestName.equals("boolean")) {
             dataModel.put( "boolean1", TemplateBooleanModel.FALSE);
             dataModel.put( "boolean2", TemplateBooleanModel.TRUE);
             dataModel.put( "boolean3", TemplateBooleanModel.TRUE);
@@ -246,14 +262,28 @@ public class TemplateTestCase extends FileTestCase {
             dataModel.put( "hash2", new BooleanHash2() );
         }
         
-        else if (testName.equals("dateformat")) {
+        else if (simpleTestName.startsWith("dateformat")) {
             GregorianCalendar cal = new GregorianCalendar(2002, 10, 15, 14, 54, 13);
             cal.setTimeZone(TimeZone.getTimeZone("GMT"));
             dataModel.put("date", new SimpleDate(cal.getTime(), TemplateDateModel.DATETIME));
             dataModel.put("unknownDate", new SimpleDate(cal.getTime(), TemplateDateModel.UNKNOWN));
+            dataModel.put("javaGMT02", TimeZone.getTimeZone("GMT+02"));
+            dataModel.put("javaUTC", TimeZone.getTimeZone("UTC"));
+            dataModel.put("adaptedToStringScalar", new Object() {
+                public String toString() {
+                    return "GMT+02";
+                }
+            });
+            dataModel.put("sqlDate", new java.sql.Date(1273955885023L));
+            dataModel.put("sqlTime", new java.sql.Time(74285023L));
         }
-    
-        else if (testName.equals("number-format")) {
+
+        else if (templateName.equals("list.ftl")
+                || templateName.equals("list2.ftl") || templateName.equals("list3.ftl")) {
+            dataModel.put("listables", new Listables());
+        }
+        
+        else if (simpleTestName.startsWith("number-format")) {
             dataModel.put("int", new SimpleNumber(new Integer(1)));
             dataModel.put("double", new SimpleNumber(new Double(1.0)));
             dataModel.put("double2", new SimpleNumber(new Double(1 + 1e-15)));
@@ -262,37 +292,74 @@ public class TemplateTestCase extends FileTestCase {
             dataModel.put("bigDecimal", new SimpleNumber(java.math.BigDecimal.valueOf(1)));
             dataModel.put("bigDecimal2", new SimpleNumber(java.math.BigDecimal.valueOf(1, 16)));
         }
+        
+        else if (simpleTestName.equals("simplehash-char-key")) {
+            HashMap mStringC = new HashMap();
+            mStringC.put("c", "string");
+            dataModel.put("mStringC", mStringC);
+            
+            HashMap mStringCNull = new HashMap();
+            mStringCNull.put("c", null);
+            dataModel.put("mStringCNull", mStringCNull);
+            
+            HashMap mCharC = new HashMap();
+            mCharC.put(Character.valueOf('c'), "char");
+            dataModel.put("mCharC", mCharC);
+            
+            HashMap mCharCNull = new HashMap();
+            mCharCNull.put("c", null);
+            dataModel.put("mCharCNull", mCharCNull);
+            
+            HashMap mMixed = new HashMap();
+            mMixed.put(Character.valueOf('c'), "char");
+            mMixed.put("s", "string");
+            mMixed.put("s2", "string2");
+            mMixed.put("s2n", null);
+            dataModel.put("mMixed", mMixed);
+        }
     
-        else if (testName.equals("default-xmlns")) {
+        else if (simpleTestName.equals("default-xmlns")) {
             InputSource is = new InputSource(getClass().getResourceAsStream("models/defaultxmlns1.xml"));
             NodeModel nm = NodeModel.parse(is);
             dataModel.put("doc", nm);
         }
         
-        else if (testName.equals("multimodels")) {
+        else if (simpleTestName.equals("multimodels")) {
             dataModel.put("test", "selftest");
             dataModel.put("self", "self");
             dataModel.put("zero", new Integer(0));
             dataModel.put("data", new MultiModel1());
         }
         
-        else if (testName.equals("stringbimethods")) {
+        else if (simpleTestName.equals("stringbimethods")) {
             dataModel.put("multi", new TestBoolean());
         }
         
-        else if (testName.equals("type-builtins")) {
+        else if (simpleTestName.startsWith("type-builtins")) {
             dataModel.put("testmethod", new TestMethod());
             dataModel.put("testnode", new TestNode());
             dataModel.put("testcollection", new SimpleCollection(new ArrayList()));
+            dataModel.put("testcollectionEx", DefaultNonListCollectionAdapter.adapt(new HashSet(), null));
+            dataModel.put("bean", new TestBean());
+        }
+
+        else if (simpleTestName.equals("date-type-builtins")) {
+            GregorianCalendar cal = new GregorianCalendar(2003, 4 - 1, 5, 6, 7, 8);
+            cal.setTimeZone(TimeZone.getTimeZone("UTC"));
+            Date d = cal.getTime();
+            dataModel.put("unknown", d);
+            dataModel.put("timeOnly", new java.sql.Time(d.getTime()));
+            dataModel.put("dateOnly", new java.sql.Date(d.getTime()));
+            dataModel.put("dateTime", new java.sql.Timestamp(d.getTime()));
         }
         
-        else if (testName.equals("var-layers")) {
+        else if (simpleTestName.equals("var-layers")) {
             dataModel.put("x", new Integer(4));
             dataModel.put("z", new Integer(4));
             conf.setSharedVariable("y", new Integer(7));
         }
         
-        else if (testName.equals("xml-fragment")) {
+        else if (simpleTestName.equals("xml-fragment")) {
             DocumentBuilderFactory f = DocumentBuilderFactory.newInstance();
             f.setNamespaceAware(true);
             DocumentBuilder db = f.newDocumentBuilder();
@@ -300,30 +367,35 @@ public class TemplateTestCase extends FileTestCase {
             dataModel.put("node", NodeModel.wrap(doc.getDocumentElement().getFirstChild().getFirstChild()));
         }
         
-        else if (testName.equals("xmlns1")) {
+        else if (simpleTestName.equals("xmlns1")) {
             InputSource is = new InputSource(getClass().getResourceAsStream("models/xmlns.xml"));
             NodeModel nm = NodeModel.parse(is);
             dataModel.put("doc", nm);
         }
         
-        else if (testName.equals("xmlns2")) {
+        else if (simpleTestName.equals("xmlns2")) {
             InputSource is = new InputSource(getClass().getResourceAsStream("models/xmlns2.xml"));
             NodeModel nm = NodeModel.parse(is);
             dataModel.put("doc", nm);
         }
         
-        else if (testName.equals("xmlns3") || testName.equals("xmlns4")) {
+        else if (simpleTestName.equals("xmlns3") || simpleTestName.equals("xmlns4")) {
             InputSource is = new InputSource(getClass().getResourceAsStream("models/xmlns3.xml"));
             NodeModel nm = NodeModel.parse(is);
             dataModel.put("doc", nm);
         }
-        else if (testName.equals("xmlns5")) {
+        else if (simpleTestName.equals("xmlns5")) {
             InputSource is = new InputSource(getClass().getResourceAsStream("models/defaultxmlns1.xml"));
             NodeModel nm = NodeModel.parse(is);
             dataModel.put("doc", nm);
         }
+        else if (simpleTestName.equals("xml-ns_prefix-scope")) {
+            InputSource is = new InputSource(getClass().getResourceAsStream("models/xml-ns_prefix-scope.xml"));
+            NodeModel nm = NodeModel.parse(is);
+            dataModel.put("doc", nm);
+        }
         
-        else if (testName.startsWith("sequence-builtins-with-")) {
+        else if (simpleTestName.startsWith("sequence-builtins")) {
             Set abcSet = new TreeSet();
             abcSet.add("a");
             abcSet.add("b");
@@ -343,38 +415,34 @@ public class TemplateTestCase extends FileTestCase {
             dataModel.put("listWithNullsOnly", listWithNullsOnly);
             
             dataModel.put("abcCollection", new SimpleCollection(abcSet));
+            
+            Set set = new HashSet();
+            set.add("a");
+            set.add("b");
+            set.add("c");
+            dataModel.put("set", set);
         }
         
-        else if (testName.equals("iso8601")) {
-            dataModel.put("javaGMT02", TimeZone.getTimeZone("GMT+02"));
-            dataModel.put("javaUTC", TimeZone.getTimeZone("UTC"));
-            dataModel.put("adaptedToStringScalar", new Object() {
-                public String toString() {
-                    return "GMT+02";
-                }
-            });
-        }
-        
-        else if (testName.equals("number-to-date")) {
+        else if (simpleTestName.equals("number-to-date")) {
           dataModel.put("bigInteger", new BigInteger("1305575275540"));
           dataModel.put("bigDecimal", new BigDecimal("1305575275539.5"));
         }
         
-        else if (testName.equals("varargs")) {
+        else if (simpleTestName.equals("varargs")) {
           dataModel.put("m", new VarArgTestModel());
         }
         
-        else if (testName.startsWith("overloaded-methods-")) {
+        else if (simpleTestName.startsWith("overloaded-methods-") && !simpleTestName.startsWith("overloaded-methods-2-")) {
           dataModel.put("obj", new OverloadedMethods());
         }
         
-        else if (testName.startsWith("boolean-formatting")) {
+        else if (simpleTestName.startsWith("boolean-formatting")) {
           dataModel.put("beansBoolean", new BooleanModel(Boolean.TRUE, (BeansWrapper) conf.getObjectWrapper()));
           dataModel.put("booleanAndString", new BooleanAndStringTemplateModel());
           dataModel.put("booleanVsStringMethods", new BooleanVsStringMethods());
         }
         
-        else if (testName.startsWith("number-math-builtins")) {
+        else if (simpleTestName.startsWith("number-math-builtins")) {
             dataModel.put("fNan", Float.valueOf(Float.NaN));
             dataModel.put("dNan", Double.valueOf(Double.NaN));
             dataModel.put("fNinf", Float.valueOf(Float.NEGATIVE_INFINITY));
@@ -399,40 +467,55 @@ public class TemplateTestCase extends FileTestCase {
             dataModel.put("bdp", BigDecimal.valueOf(0.05));
           }
           
-        else if (testName.startsWith("classic-compatible")) {
+        else if (simpleTestName.startsWith("classic-compatible")) {
             dataModel.put("array", new String[] { "a", "b", "c" });
             dataModel.put("beansArray", new BeansWrapper().wrap(new String[] { "a", "b", "c" }));
             dataModel.put("beanTrue", new BeansWrapper().wrap(Boolean.TRUE));
             dataModel.put("beanFalse", new BeansWrapper().wrap(Boolean.FALSE));
         }
+        
+        else if (simpleTestName.startsWith("overloaded-methods-2-")) {
+            dataModel.put("obj", new OverloadedMethods2());
+            final boolean dow = conf.getObjectWrapper() instanceof DefaultObjectWrapper;
+            dataModel.put("dow", dow);
+            dataModel.put("dowPre22", dow
+                    && ((DefaultObjectWrapper) conf.getObjectWrapper()).getIncompatibleImprovements()
+                            .intValue() < _TemplateAPI.VERSION_INT_2_3_22);
+        }
     }
     
-    public void runTest() {
+    public void runTest() throws IOException {
+        Template template;
         try {
             template = conf.getTemplate(templateName);
-        } catch (Exception e) {
-            StringWriter sw = new StringWriter();
-            e.printStackTrace(new PrintWriter(sw));
-            fail("Could not load template " + StringUtil.jQuote(templateName) + "\n" + sw.toString());
+        } catch (IOException e) {
+            throw new AssertionFailedError(
+                    "Could not load template " + StringUtil.jQuote(templateName) + ":\n" + getStackTrace(e));
         }
+        ASTPrinter.validateAST(template);
         
         StringWriter out = noOutput ? null : new StringWriter();
         try {
             template.process(dataModel, out != null ? out : NullWriter.INSTANCE);
-        } catch (Exception e) {
-            StringWriter sw = new StringWriter();
-            e.printStackTrace(new PrintWriter(sw));
-            fail("Could not process template " + templateName + "\n" + sw.toString());
+        } catch (TemplateException e) {
+            throw new AssertionFailedError("Template " + StringUtil.jQuote(templateName) + " has stopped with error:\n"
+                        + getStackTrace(e));
         }
         
         if (out != null) {
             assertExpectedFileEqualsString(getName(), out.toString());
         }
     }
-    
+
+    private String getStackTrace(Throwable e) {
+        StringWriter sw = new StringWriter();
+        e.printStackTrace(new PrintWriter(sw));
+        return sw.toString();
+    }
+
     @Override
     protected File getExpectedFileDirectory() throws IOException {
-        return new File(super.getExpectedFileDirectory(), "references");
+        return new File(super.getExpectedFileDirectory(), "expected");
     }
 
     @Override
@@ -484,16 +567,16 @@ public class TemplateTestCase extends FileTestCase {
       }
     }
 
-   public Object getTestBean()
+   public Object getTestMapBean()
     {
-        Map testBean = new TestBean();
+        Map testBean = new TestMapBean();
         testBean.put("name", "Chris");
         testBean.put("location", "San Francisco");
         testBean.put("age", new Integer(27));
         return testBean;
     }
 
-    public static class TestBean extends HashMap {
+    public static class TestMapBean extends HashMap {
         public String getName() {
             return "Christopher";
         }
@@ -501,5 +584,21 @@ public class TemplateTestCase extends FileTestCase {
             return 7;
         }
     }
+
+    public static class TestBean {
+
+        public int m(int n) {
+            return n * 10;
+        }
+
+        public int mOverloaded(int n) {
+            return n * 10;
+        }
+
+        public String mOverloaded(String s) {
+            return s.toUpperCase();
+        }
+        
+    }
     
 }
